#!/usr/bin/perl -w

$nvalues = q{ adeilnorst_};
$cvalues = q{      9876543210};
$dvalues = q{ ABCDEFGHIJKLMNO};
$evalues = q{ z  @ZYXWVUTSRQP};
$fvalues = q{ bcfghjkmpquvwxy};

%enumorder = ( X_SYSFLAG => 0x55,
	       X_SCANCODE => 0x56,
	       X_DATATYPE => 0x57,
	       X_KEY => 0x58,
	       X_KEY_FB => 0x59,
	       X_KEY_FC => 0x5A );

sub pack_sym_string {
  my $text = shift;
  my $hex = '';
  my $data = '';
  my $v;

  foreach my $c (split //, $text) {
    if (($v = index $nvalues, $c) != -1) {
      $hex .= sprintf '%X', $v;
    }
    elsif (($v = index $cvalues, $c) != -1) {
      $hex .= sprintf 'C%X', $v;
    }
    elsif (($v = index $dvalues, $c) != -1) {
      $hex .= sprintf 'D%X', $v;
    }
    elsif (($v = index $evalues, $c) != -1) {
      $hex .= sprintf 'E%X', $v;
    }
    elsif (($v = index $fvalues, $c) != -1) {
      $hex .= sprintf 'F%X', $v;
    }
  }

  $hex .= '0';
  $hex =~ s/([0-9A-F][0-9A-F]?)/0$1h,/g;
  $hex =~ s/,$//;
  return $hex;
}

sub process_table_file {
  my ($fname, $init, $tab, $ent) = @_;
  my %symvalues = ();
  my %symcat = ();

  open IN, "$fname" or die "can't open $fname";

  while (<IN>) {
    chomp;
    s/[#;].*//;
    if (/^$init([A-Za-z0-9_]+)\s*=\s*([0-9A-Fa-f]+)/) {
      $symvalues{$1} = hex $2;
    }
    elsif (/^([A-F0-9]+):([A-Za-z0-9_]+)\s*=\s*([0-9A-Fa-f]+)/) {
      $symvalues{$2} = (hex $3) + 256 * (hex $1);
      $symcat{$2} = $1 . 'h';
    }
    elsif (/^(X_[A-Z0-9_]+):([A-Za-z0-9_]+)\s*=\s*([0-9A-Fa-f]+)/) {
      $symvalues{$2} = (hex $3) + 256 * $enumorder{$1};
      $symcat{$2} = $1;
    }
    elsif (/\S/) {
      print STDERR "unable to parse '$_'\n";
    }
  }
  close IN;

  my @byaddr = sort { $symvalues{$a} <=> $symvalues{$b} } keys %symvalues;
  my @byname = sort { (lc $a) cmp (lc $b) } keys %symvalues;

  print ASMDATA "${tab}ByValueTable:\n";
  foreach my $sym (@byaddr) {
    print ASMDATA "\tdw ${ent}_${sym}\n";
  }
  print ASMDATA "${tab}ByValueTableSize equ \$ - ${tab}ByValueTable\n\n";

  print ASMDATA "\tdb -1,-1,0\n";
  print ASMDATA "${tab}Names:\n";
  foreach my $sym (@byname) {
    if (defined $symcat{$sym}) {
      printf ASMDATA ("${ent}_${sym}: \tdb 0%02Xh, ",
		  ($symvalues{$sym} % 256));
      print ASMDATA $symcat{$sym}, ", ";
    }
    else {
      printf ASMDATA ("${ent}_${sym}: \tdb 0%02Xh,0%02Xh, ",
		  ($symvalues{$sym} % 256),
		  ($symvalues{$sym} / 256));
    }
    print ASMDATA pack_sym_string($sym), "\n";
  }
  print ASMDATA "${tab}NamesEnd:\n\n";

  print BYVALUE "static const builtin_symbol ${tab}ByValueTable[] = {\n";
  for ($i = 0; $i <= $#byaddr; $i++) {
    my $sym = $byaddr[$i];
    print BYVALUE '{', $symvalues{$sym}, ', "', $sym, '"}';
    print BYVALUE ',' if ($i < $#byaddr);
    print BYVALUE "\n";
  }
  print BYVALUE "};\n";

  print BYNAME "static const builtin_symbol ${tab}ByNameTable[] = {\n";
  for ($i = 0; $i <= $#byname; $i++) {
    my $sym = $byname[$i];
    print BYNAME '{', $symvalues{$sym}, ', "', lc $sym, '"}';
    print BYNAME ',' if ($i < $#byname);
    print BYNAME "\n";
  }
  print BYNAME "};\n";
}

open ASMDATA, ">sysdata.asm" or die "can't open sysdata.asm";
print ASMDATA "; AUTOMATICALLY GENERATED - DO NOT EDIT\n\n";

my $infostruct = '
typedef struct _builtin_symbol {
  unsigned int value;
  const char *name;
} builtin_symbol;';

open BYVALUE, ">../tools/byvalue.h" or die "can't open byvalue.h";
print BYVALUE "/* AUTOMATICALLY GENERATED - DO NOT EDIT */\n$infostruct\n\n";

open BYNAME, ">../tools/byname.h" or die "can't open byname.h";
print BYNAME "/* AUTOMATICALLY GENERATED - DO NOT EDIT */\n$infostruct\n\n";

process_table_file("romcalls.dat", "_", "romCall", "rc");
process_table_file("sysaddrs.dat", "", "sysAddr", "sa");
process_table_file("enums.dat", "", "sysEnum", "en");
process_table_file("flags.dat", "", "sysFlag", "fl");

close ASMDATA;
close BYVALUE;
close BYNAME;
