#!/usr/bin/perl -w

%rulecmds = ( ixpfx => "R_IX",
	      __ => "R_END",
	      byte => "R_BYTE",
	      word => "R_WORD",
	      rel => "R_REL",
	      sbyte => "R_SBYTE",
	      sreg => "R_SREG",
	      dreg => "R_DREG",
	      cond => "R_DREG",
	      bitreg => "R_BITREG",
	      rst => "R_RST" );

%literalreg = ( B => "X_REG_B", C => "X_REG_C",
		D => "X_REG_D", E => "X_REG_E",
		H => "X_REG_H", L => "X_REG_L",
		iHL => "X_REG_iHL", A => "X_REG_A",
		BC => "X_REG_BC", DE => "X_REG_DE",
		HL => "X_REG_HL", SP => "X_REG_SP",
		IX => "X_REG_IX", IY => "X_REG_IY",
		"AF'" => "X_REG_AF2", AF => "X_REG_AF",
		I => "X_REG_I", R => "X_REG_R",
		IXH => "X_REG_IXH", IXL => "X_REG_IXL",
		IYH => "X_REG_IYH", IYL => "X_REG_IYL" );

%regcat = ( bcdehlma => "AC_BCDEHLMA",
	    bcdehla => "AC_BCDEHLA",
	    bcdea => "AC_BCDEA",
	    bdhs => "AC_BDHS",
	    bdha => "AC_BDHA",
	    bds => "AC_BDS",
	    IXhl => "AC_IXHL",
	    IYhl => "AC_IYHL",
	    zcps => "AC_ZCPS",
	    zc => "AC_ZC" );

%flagdesc = ( '----' => 'IC_NONE',
              '---1' => 'IC_NONE',
              '---c' => 'IC_NONE',
              '--l-' => 'IC_BLOCK',
              '?b??' => 'IC_BLOCKIO',
              'szi-' => 'IC_INTSTATE',
              'szl-' => 'IC_BLOCK',
              'szp-' => 'IC_PARITY',
              'szp0' => 'IC_PARITY_C0',
              'szpc' => 'IC_PARITY',
              'szv-' => 'IC_OVERFLOW',
              'szvc' => 'IC_OVERFLOW' );

%output_rules = ();

sub gen_symbol {
  my ($instname, $argpat) = @_;

  $argpat =~ s/\bbcdehlma\b/r/g;
  $argpat =~ s/\bbcdehla\b/s/g;
  $argpat =~ s/\bbcdea\b/t/g;
  $argpat =~ s/\bbdhs\b/rr/g;
  $argpat =~ s/\bbdha\b/ss/g;
  $argpat =~ s/\bbds\b/tt/g;
  $argpat =~ s/\bIXhl\b/IXr/g;
  $argpat =~ s/\bIYhl\b/IYr/g;
  $argpat =~ s/\bzcps\b/ccc/g;
  $argpat =~ s/\bzc\b/cc/g;
  $argpat =~ s/\$/z/g;
  $argpat =~ s/\*/n/g;
  $argpat =~ s/\((.*?)\)/i$1/g;
  $argpat =~ s/'//g;
  $argpat =~ s/\+/p/g;
  $argpat =~ s/,/_/g;
  $argpat =~ s/ *//g;

  if ($argpat ne "") {
    return $instname . "_" . $argpat;
  }
  elsif ($instname ne "") {
    return $instname;
  }
  else {
    return "null";
  }
}

sub get_template {
  my $argpat = shift;
  my @result = ();

  $argpat =~ s/ *//g;

  if ($argpat =~ /^\((.*)\)$/) {
    push @result, "X_PAREN";
    push @result, get_template($1);
  }
  elsif ($argpat =~ /^(.*)\+(.*)$/) {
    push @result, "X_ADD";
    push @result, get_template($1);
    push @result, get_template($2);
  }
  elsif (defined($literalreg{$argpat})) {
    push @result, "X_REGVAL";
    push @result, $literalreg{$argpat};
  }
  elsif ($argpat eq "0" or $argpat eq "1" or $argpat eq "2") {
    push @result, "X_DEC6+$argpat";
  }
  else {
    if ($argpat =~ s/\$([0-9]+)$//g) {
      $curargnum = $1;
    }

    if ($argpat eq "*") {
      push @result, "X_ARGVAL";
      push @result, "X_DEC6+AC_CONSTANT+" . ($curargnum - 1);
    }
    elsif (!defined($regcat{$argpat})) {
      push @result, "???";
      print STDERR "undefined category $argpat\n";
    }
    else {
      push @result, "X_REGVAL";
      push @result, "X_ARGVAL";
      push @result, "X_DEC6+" . $regcat{$argpat} . "+" . ($curargnum - 1);
    }
    $curargnum++;
  }

  return @result;
}

sub print_template {
  my $tsym = shift;

  print TEMPLATES "templ_$tsym: db ";
  print TEMPLATES join(",", ($#_ + 2), @_);
  print TEMPLATES "\n";

  print CTEMPLATES "static const unsigned char templ_$tsym\[] = {";
  print CTEMPLATES join(",", ($#_ + 1), @_);
  print CTEMPLATES "};\n";
}

sub add_forms {
  my ($mnemonic, $argpat, $rule, $ireg, $iprefix, $tstates, $flags) = @_;

  $argpat =~ s/I%/$ireg/g;
  $rule =~ s/\bprefix\b/$iprefix/g;
  $rule .= ' __';

  $bytemnemonic[$byteval] = $mnemonic;
  $sym = gen_symbol($mnemonic, $argpat);
  $tsym = gen_symbol("", $argpat);
  $sym =~ s/z[0-9]+//g;

  $bytesyms[$byteval] = $sym;
  $bytetemplsyms[$byteval] = $tsym;

  print CODES "I_$sym equ $byteval\n";
  print CCODES "#define I_$sym $byteval\n";

  $output_rules{$sym} = $rule;

  $curargnum = 1;
  @templ = ();
  for $ap (split m{,}, $argpat) {
    push @templ, get_template($ap);
  }

  if (!$templsymswritten{$tsym}) {
    print_template($tsym, @templ);
    $templsymswritten{$tsym} = 1;
  }

  if ($tstates =~ /^(\d+)i(\d+)$/) {
    die if $1 < 4 or $1 > (7 + 4);
    die if $2 < 4 or $2 > (15 + 4);
    $bytetstates[$byteval] = ($1 - 4) * 16 + ($2 - 4);
  }
  elsif ($tstates =~ /^(\d+)c(\d+)$/) {
    die if $1 < 4 or $1 > (15 + 4);
    die if $2 < ($1 + 5) or $2 > ($1 + 7);
    $bytetstates[$byteval] = 0x80 + ($2 - $1 - 5) * 32 + ($1 - 4);
  }
  elsif ($tstates =~ /^\d+$/) {
    die if $tstates < 4 or $tstates > (31 + 4);
    $bytetstates[$byteval] = 0xe0 + $tstates - 4;
  }

  die if !defined $flagdesc{$flags};
  $flagmask = $flags;
  $flagmask =~ s/[^-]/1/g;
  $flagmask =~ s/-/0/g;
  $byteflags[$byteval] = "$flagdesc{$flags}+${flagmask}b";
}


open INSTINFO, "insts.dat" or die "can't open insts.dat";

open CODES, ">insts.inc" or die "can't open insts.inc";
open RULES, ">irules.asm" or die "can't open irules.asm";
open TEMPLATES, ">itempl.asm" or die "can't open itempl.asm";
open TABLES, ">itables.asm" or die "can't open itables.asm";
open INFOTABLES, ">../plugins/iinfo.asm" or die "can't open iinfo.asm";
open CCODES, ">../tools/insts.h" or die "can't open insts.h";
open CTEMPLATES, ">../tools/itempl.h" or die "can't open itempl.h";

print CODES "; AUTOMATICALLY GENERATED - DO NOT EDIT\n\n";
print RULES "; AUTOMATICALLY GENERATED - DO NOT EDIT\n\n";
print TEMPLATES "; AUTOMATICALLY GENERATED - DO NOT EDIT\n\n";
print TABLES "; AUTOMATICALLY GENERATED - DO NOT EDIT\n\n";
print INFOTABLES "; AUTOMATICALLY GENERATED - DO NOT EDIT\n\n";
print CCODES "/* AUTOMATICALLY GENERATED - DO NOT EDIT */\n\n";
print CTEMPLATES "/* AUTOMATICALLY GENERATED - DO NOT EDIT */\n\n";

print_template('null');
print_template('_n_n', 'X_ARGVAL', 'X_DEC6+AC_CONSTANT+0',
	       'X_ARGVAL', 'X_DEC6+AC_CONSTANT+1');

$byteval = 1;
@bytesyms = ("null");
@bytetemplsyms = ("null");
@bytemnemonic = ("xxx");
@bytetstates = (0);
@byteflags = (0);
%templsymswritten = ("null" => 1);
%output_rules = ("null" => "__");

while (<INSTINFO>) {
  s/[#;].*//;
  chomp;

  @info = split /\t/;
  if (@info >= 3) {
    $mnemonic = $info[0];
    $argpat = $info[1];
    $rule = $info[2];
    $tstates = $info[3];
    $flags = $info[4];

    if ($rule =~ /\bprefix\b/) {
      add_forms($mnemonic, $argpat, $rule, "IX", "ixpfx FD", $tstates, $flags);
      $byteval++;
      add_forms($mnemonic, $argpat, $rule, "IY", "FD", $tstates, $flags);
      $byteval++;
    }
    else {
      add_forms($mnemonic, $argpat, $rule, "??", "??", $tstates, $flags);
      $byteval++;
    }
  }
}

@ordered_rules = sort { (reverse $output_rules{$a})
			  cmp (reverse $output_rules{$b}) }
  (keys %output_rules);

@ordered_ruletext = ();
$nextrule = $nextsym = 'fnord';

for $sym (@ordered_rules) {
  $rule = $fullrule = $output_rules{$sym};

  if ($rule =~ /^(.*?) *$nextrule$/) {
    $rule = $1;

    printf STDERR "Merging %-16s ( %s )\n", $sym, $fullrule;
    printf STDERR "   with %-16s ( %s )\n", $nextsym, $nextrule;
  }

  $nextrule = $fullrule;
  $nextsym = $sym;

  $ruletext = "rule_$sym: db ";
  for $rb (split / +/, $rule) {
    if ($rulecmds{$rb}) {
      $ruletext .= $rulecmds{$rb} . ",";
    }
    elsif ($rb =~ /^[0-9A-F][0-9A-F]$/) {
      $ruletext .= "0${rb}h,";
    }
    else {
      print STDERR "unknown rule command $rb\n";
      exit 1;
    }
  }

  $ruletext =~ s/,$//;
  $ruletext =~ s/db $//;

  unshift @ordered_ruletext, $ruletext;
}

for $text (@ordered_ruletext) {
  print RULES $text, "\n";
  print RULES "\n" if ($text =~ /R_END$/);
}

$byteval--;
print CODES "I_last equ $byteval\n";
print CCODES "#define I_last $byteval\n";

close INSTINFO;

print TABLES "normalInstructionRuleLTable:";
for ($i = 0; $i < 256; $i++) {
  print TABLES (($i % 4) ? "," : "\n\tdb ");
  if ($i <= $#bytesyms) {
    print TABLES "low(rule_$bytesyms[$i])";
  }
  else {
    print TABLES "low(rule_null)";
  }
}
print TABLES "\nnormalInstructionRuleHTable:";
for ($i = 0; $i < 256; $i++) {
  print TABLES (($i % 4) ? "," : "\n\tdb ");
  if ($i <= $#bytesyms) {
    print TABLES "high(rule_$bytesyms[$i])";
  }
  else {
    print TABLES "high(rule_null)";
  }
}
print TABLES "\nnormalInstructionTemplateLTable:";
for ($i = 0; $i < 256; $i++) {
  print TABLES (($i % 4) ? "," : "\n\tdb ");
  if ($i <= $#bytesyms) {
    print TABLES "low(templ_$bytetemplsyms[$i])";
  }
  else {
    print TABLES "low(templ_null)";
  }
}
print TABLES "\nnormalInstructionTemplateHTable:";
for ($i = 0; $i < 256; $i++) {
  print TABLES (($i % 4) ? "," : "\n\tdb ");
  if ($i <= $#bytesyms) {
    print TABLES "high(templ_$bytetemplsyms[$i])";
  }
  else {
    print TABLES "high(templ_null)";
  }
}
print TABLES "\nnormalInstructionMnemonicTable:";
for ($i = 0; $i < 256; $i++) {
  print TABLES (($i % 4) ? "," : "\n\tdb ");
  if ($i <= $#bytesyms) {
    print TABLES "mnem_$bytemnemonic[$i]";
  }
  else {
    print TABLES "mnem_xxx";
  }
}
print TABLES "\n";

print INFOTABLES "normalInstructionTStateTable:";
for ($i = 0; $i <= $#bytetstates; $i++) {
  print INFOTABLES (($i % 4) ? "," : "\n\tdb ");
  print INFOTABLES $bytetstates[$i];
}
print INFOTABLES "\nnormalInstructionFlagInfoTable:";
for ($i = 0; $i <= $#byteflags; $i++) {
  print INFOTABLES (($i % 4) ? "," : "\n\tdb ");
  print INFOTABLES $byteflags[$i];
}
print INFOTABLES "\n";

print CTEMPLATES "\nstatic const char * const normal_mnemonics[] = {";
for ($i = 0; $i < 255; $i++) {
  print CTEMPLATES "\n" if (($i % 8) == 0);
  if ($i <= $#bytesyms) {
    print CTEMPLATES '"', lc $bytemnemonic[$i], '",';
  }
  else {
    print CTEMPLATES "NULL,";
  }
}
print CTEMPLATES "NULL };\n\n";
print CTEMPLATES "static const unsigned char * const normal_templates[] = {";
for ($i = 0; $i < 255; $i++) {
  print CTEMPLATES "\n" if (($i % 4) == 0);
  if ($i <= $#bytesyms) {
    print CTEMPLATES "templ_$bytetemplsyms[$i],";
  }
  else {
    print CTEMPLATES "NULL,";
  }
}
print CTEMPLATES "NULL };\n";

close CODES;
close RULES;
close TEMPLATES;
close TABLES;
close INFOTABLES;
close CCODES;
close CTEMPLATES;
